// Title:   CSG_Tool.cpp
// Created: Thu Sep 25 15:07:33 2003
// Authors: Richard Keiser, Mark Pauly
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include <qwhatsthis.h>
#include <qapplication.h>
#include <qradiobutton.h>
#include <qspinbox.h>
#include <qpainter.h>
#include <qcheckbox.h>
#include <qmessagebox.h>
#include <qlabel.h>

#include "../../../../Core/Scene/src/Object.h"
#include "../../../../Core/Scene/src/Scene.h"
#include "../../../../Core/Configuration/src/Configuration.h"
#include "../../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../../Core/DataStructures/src/ClusterNode.h"
#include "../../../../Core/CoreTools/src/CoreTools.h"
#include "../../../../Core/Interfaces/src/ToolInterface.h"
#include "../../../../Core/Transformer/src/Transformer.h"
#include "../../../../Core/ToolManager/src/ToolManager.h"

#include "../../../../Utilities/src/Common.h"
#include "../../../../Utilities/UserMessages/src/StatusBar.h"
#include "../../../../Utilities/src/TrackBall.h"

#ifndef MY_PI
#define MY_PI 3.14159265358979323846f
#endif

#include "../UI/CSG_ConfigurationWidget.h" 
#include "CSG_Tool.h"


#define OP_A_union_B 0
#define OP_A_intersection_B 1
#define OP_A_minus_B 2
#define OP_B_minus_A 3

// ******************
// 'What's this' text
// ******************

static const char *whatsThisText = "<img source=\"CSG_\"> "
"Click this button to perform a CSG operation on two models";

CSG_Tool::CSG_Tool() {

	standardCursor = new QCursor (CrossCursor);
	transformCursor = new QCursor (SizeAllCursor);
	toolCursor     = standardCursor;
	usePreviewRenderer = true;	
	
	configurationWidget = new CSG_ConfigurationWidget(0);
}


CSG_Tool::~CSG_Tool() {

	delete standardCursor;
	delete transformCursor;
	delete configurationWidget;
	deselectTool();

}

// **************
// public methods
// **************

QToolButton *CSG_Tool::addButtonToToolBar (QToolBar *toolBar) {


	QIconSet iconSet;
	QPixmap  normalIcon ("Resources/CSGToolIcon.png"),
		     disabledIcon ("Resources/CSGToolIcon_disabled.png");

	iconSet.setPixmap (normalIcon, QIconSet::Automatic, QIconSet::Normal);
	iconSet.setPixmap (disabledIcon, QIconSet::Automatic, QIconSet::Disabled);
	csg_Button = new QToolButton (iconSet, "CSG Tool", "Perform boolean operations",
	                                0, 0, toolBar, "CSG Tool");
	csg_Button->setToggleButton (true);
	csg_Button->setEnabled (this->checkToolAvailability());
	QWhatsThis::add (csg_Button, whatsThisText);
	QMimeSourceFactory::defaultFactory()->setPixmap ("csg_", normalIcon);

	operation = -1;		// no selected CSG operation by default

	configurationWidget->previewRendererCheckBox->setChecked (usePreviewRenderer);

	return csg_Button;
}


QCursor *CSG_Tool::getToolCursor() const {
	return toolCursor;
}


QToolButton* CSG_Tool::getToolButton () {

	return csg_Button;
}



QWidget *CSG_Tool::getToolConfigurationWidget() const {
	return configurationWidget;
}


bool CSG_Tool::isAvailable() const {
	return this->checkToolAvailability();
} 



void CSG_Tool::handleMouseMoveEvent (QMouseEvent *me) {
	
	float                                normalizedMousePositions[4];
	float                                rotationVector[4],
		                                 sx,
		                                 sy,
		                                 sz;
	QSize                                viewPortSize;
	Scene                                *scene;
	Vector3D                             transformation,
		                                 finalTransformation;

	ToolInterface::handleMouseMoveEvent(me);

	viewPortSize = Configuration::getInstance()->getViewPortSize();
	scene = Scene::getInstance();

	if (!isGlobalTransformationActive()) {
		
		// left mouse button
		if (me->state() & (QMouseEvent::LeftButton | QMouseEvent::ShiftButton)) {
			
			// rotate
			
			// normalize mouse positions into rectangle, [-1.0, 1.0] x [-1.0, 1.0]
			normalizedMousePositions[0] = (float)lastMouseDragPosition.x() / (float)viewPortSize.width()  * 2.0f - 1.0f;
			normalizedMousePositions[1] = (float)lastMouseDragPosition.y() / (float)viewPortSize.height() * 2.0f - 1.0f;
			normalizedMousePositions[2] = (float)me->x() / (float)viewPortSize.width()  * 2.0f - 1.0f;
			normalizedMousePositions[3] = (float)me->y() / (float)viewPortSize.height() * 2.0f - 1.0f;
			
			if (TrackBall::calculateRotationMatrix (normalizedMousePositions, rotationVector) == true) {			
				scene->getActiveObject()->rotate (rotationVector[0] * PI_180, rotationVector[1], rotationVector[2], rotationVector[3]);			
			}		
		}
		
		// middle mouse button
		else if ((me->state() == QMouseEvent::MidButton)) {
			
			// translate in z direction
			
			transformation[0] = 0.0f;
			transformation[1] = 0.0f;
			transformation[2] = -2.0f * (me->y() - lastMouseDragPosition.y());
			
			scene->getActiveObject()->translate (transformation);
			
		}
		
		// right mouse button
		else if (me->state() == QMouseEvent::RightButton) {
			
			// translate in x/y direction
			
			transformation[0] = me->x() - lastMouseDragPosition.x();
			transformation[1] = lastMouseDragPosition.y() - me->y();
			transformation[2] = 0.0f;
			
			scene->getActiveObject()->translate (transformation);
		}
		
		// right mouse button + CONTROL
		else if (me->state() == (QMouseEvent::RightButton | QMouseEvent::ControlButton)) {
			
			// uniform scaling in x, y and z-direction
			
			sy = 1.0f + 0.025f * (lastMouseDragPosition.y() - me->y());
			sy = (sy > 0.1f) ? sy : 0.1f;
			sx = sy;
			sz = sy;
			scene->getActiveObject()->scale (sx, sy, sz);
			
		}
		
		// right mouse button + SHIFT
		else if (me->state() == (QMouseEvent::RightButton | QMouseEvent::ShiftButton)) {
			
			// translate in z direction
			
			transformation[0] = 0.0f;
			transformation[1] = 0.0f;
			transformation[2] = -2.0f * (me->y() - lastMouseDragPosition.y());
			
			scene->getActiveObject()->translate (transformation);
		}		
			
		computeCSG();
		Scene::getInstance()->emitObjectModifiedSignal();
		Scene::getInstance()->emitRenderFrameSignal();

	}

	lastMouseDragPosition = me->pos();

}


void CSG_Tool::handleMouseReleaseEvent (QMouseEvent *me) {

	// when releasing the mouse the intersection curves are sampled
	// but only if the object has changed.
	if (!isGlobalTransformationActive()) {
		sampleIntersectionCurves();
	}

	RendererManager::getInstance()->setPreviewRendererEnabled (Configuration::getInstance()->isPreviewRendererEnabled(), true);

}



void CSG_Tool::handleMousePressEvent (QMouseEvent *me) {

	ToolInterface::handleMousePressEvent(me);

	lastMouseDragPosition = me->pos();

	// enable preview renderer while rotating etc.
	RendererManager::getInstance()->setPreviewRendererEnabled (usePreviewRenderer, true);
}

void CSG_Tool::handleKeyReleaseEvent (QKeyEvent *ke) {
	ke->ignore();
}


void CSG_Tool::handleKeyPressEvent (QKeyEvent *ke) {

	ToolInterface::handleKeyPressEvent(ke);

	if (ke->key() == Qt::Key_Asterisk) {
		
		doRefinement();
		Scene::getInstance()->emitObjectModifiedSignal();
		Scene::getInstance()->emitRenderFrameSignal();

	}
	else if (ke->key() == Qt::Key_Backspace) {
		
		handleResamplingChanged();

	}
}

void CSG_Tool::setGlobalTransformationActive (bool isOn) { 

	ToolInterface::setGlobalTransformationActive(isOn);

	if (!isOn) {
		toolCursor = standardCursor;
	}
	else {
		toolCursor = transformCursor;
	}

	RendererManager *rendererManager;
	QWidget         *rendererWidget;

	rendererManager = RendererManager::getInstance();
	if (rendererManager->isPreviewRendererEnabled()) {
		rendererWidget = rendererManager->getPreviewRenderer()->getRendererWidget();
		rendererWidget->setCursor (*toolCursor);
	}
	else {
		rendererWidget = rendererManager->getMainRenderer()->getRendererWidget();
		rendererWidget->setCursor (*toolCursor);
	}
}


void CSG_Tool::selectTool() {
	
	Scene* scene;

	scene = Scene::getInstance();

	// get informed whenever the number of objects in the scene changes
	this->connect (Scene::getInstance(), SIGNAL (objectAdded()),
		           this, SLOT (handleObjectListChanged()));

	// get informed whenever the number of objects in the scene changes
	this->connect (Scene::getInstance(), SIGNAL (objectRemoved()),
		           this, SLOT (handleObjectListChanged()));

	this->connect (configurationWidget->previewRendererCheckBox, SIGNAL (toggled (bool)),
		           this, SLOT (handlePreviewRendererToggled (bool)));
	this->connect (configurationWidget->applyPushButton, SIGNAL (released()),
		           this, SLOT (handleApply ()));
	this->connect (configurationWidget->cancelPushButton, SIGNAL (released()),
		           this, SLOT (handleCancel ()));
	this->connect (configurationWidget->modelAComboBox, SIGNAL (activated(int)),
		           this, SLOT (handleComboBoxesUpdated()));
	this->connect (configurationWidget->modelBComboBox, SIGNAL (activated(int)),
		           this, SLOT (handleComboBoxesUpdated()));
	this->connect (configurationWidget->a_plus_B_RadioButton, SIGNAL (stateChanged(int)),
		           this, SLOT (handleOperationChanged()));
	this->connect (configurationWidget->a_times_B_RadioButton, SIGNAL (stateChanged(int)),
		           this, SLOT (handleOperationChanged()));
	this->connect (configurationWidget->a_minus_B_RadioButton, SIGNAL (stateChanged(int)),
		           this, SLOT (handleOperationChanged()));
	this->connect (configurationWidget->b_minus_A_RadioButton, SIGNAL (stateChanged(int)),
		           this, SLOT (handleOperationChanged()));
	this->connect (configurationWidget->rebuildPushButton, SIGNAL (released()),
		           this, SLOT (handleRebuild()));
	this->connect (configurationWidget->MaxDistanceLineEdit, SIGNAL( returnPressed() ), 
				   this, SLOT( handleResamplingChanged()));
	this->connect (configurationWidget->MinDistanceLineEdit, SIGNAL( returnPressed() ), 
				   this, SLOT( handleResamplingChanged()));
	this->connect (configurationWidget->MaxAngleSpinBox, SIGNAL( valueChanged(int) ), 
				   this, SLOT( handleResamplingChanged()));
    this->connect (configurationWidget->OverlapLineEdit, SIGNAL( returnPressed() ), 
				   this, SLOT( handleResamplingChanged() ));
	this->connect (configurationWidget->MaxAngleCheckBox, SIGNAL( stateChanged(int)),
				   this, SLOT( handleResamplingChanged() ));
	this->connect (configurationWidget->MaxFeatureDistanceCheckBox, SIGNAL( stateChanged(int)),
				   this, SLOT( handleResamplingChanged() ));
	this->connect (configurationWidget->MinFeatureDistanceCheckBox, SIGNAL( stateChanged(int)),
				   this, SLOT( handleResamplingChanged() ));

	
	featurePoints		= NULL;
	clipPositionsA		= NULL;
	clipPositionsB		= NULL;
	clipSurfelsA		= NULL;
	clipSurfelsB		= NULL;
	clipSurfelsOtherA	= NULL;
	clipSurfelsOtherB	= NULL;
	clipKdTreeA			= NULL;
	clipKdTreeB			= NULL;

	handleObjectListChanged();
	
	if (!objectA || !objectB) {
		// calls deselectTool!
		ToolManager::getInstance()->switchToNavigationTool();
	}
	else {
		initCSG();
	}

}


void CSG_Tool::deselectTool() {

	this->disconnect (Scene::getInstance(), SIGNAL (objectAdded()), this, SLOT (handleObjectListChanged()));
	this->disconnect (Scene::getInstance(), SIGNAL (objectRemoved()), this, SLOT (handleObjectListChanged()));

	this->disconnect (configurationWidget->previewRendererCheckBox, SIGNAL (toggled (bool)),
		           this, SLOT (handlePreviewRendererToggled (bool)));
	this->disconnect (configurationWidget->applyPushButton, SIGNAL (released()),
		           this, SLOT (handleApply ()));
	this->disconnect (configurationWidget->cancelPushButton, SIGNAL (released()),
		           this, SLOT (handleCancel ()));
	this->disconnect (configurationWidget->modelAComboBox, SIGNAL (activated(int)),
		           this, SLOT (handleComboBoxesUpdated()));
	this->disconnect (configurationWidget->modelBComboBox, SIGNAL (activated(int)),
		           this, SLOT (handleComboBoxesUpdated()));
	this->disconnect (configurationWidget->a_plus_B_RadioButton, SIGNAL (stateChanged(int)),
		           this, SLOT (handleOperationChanged()));
	this->disconnect (configurationWidget->a_times_B_RadioButton, SIGNAL (stateChanged(int)),
		           this, SLOT (handleOperationChanged()));
	this->disconnect (configurationWidget->a_minus_B_RadioButton, SIGNAL (stateChanged(int)),
		           this, SLOT (handleOperationChanged()));
	this->disconnect (configurationWidget->b_minus_A_RadioButton, SIGNAL (stateChanged(int)),
		           this, SLOT (handleOperationChanged()));
	this->disconnect (configurationWidget->rebuildPushButton, SIGNAL (released()),
		           this, SLOT (handleRebuild()));
	this->disconnect (configurationWidget->MaxDistanceLineEdit, SIGNAL( returnPressed() ), 
				   this, SLOT( handleResamplingChanged()));
	this->disconnect (configurationWidget->MinDistanceLineEdit, SIGNAL( returnPressed() ), 
				   this, SLOT( handleResamplingChanged()));
	this->disconnect (configurationWidget->MaxAngleSpinBox, SIGNAL( valueChanged(int) ), 
				   this, SLOT( handleResamplingChanged()));
    this->disconnect (configurationWidget->OverlapLineEdit, SIGNAL( returnPressed() ), 
				   this, SLOT( handleResamplingChanged() ));
	this->disconnect (configurationWidget->MaxAngleCheckBox, SIGNAL( stateChanged(int)),
				   this, SLOT( handleResamplingChanged() ));
	this->disconnect (configurationWidget->MaxFeatureDistanceCheckBox, SIGNAL( stateChanged(int)),
				   this, SLOT( handleResamplingChanged() ));
	this->disconnect (configurationWidget->MinFeatureDistanceCheckBox, SIGNAL( stateChanged(int)),
				   this, SLOT( handleResamplingChanged() ));

	reset();

}

// ************
// public slots
// ************
void CSG_Tool::handleRendererWidgetRepainted() {
}

// ***************
// private methods
// ***************



void CSG_Tool::computeCSG() {

	Scene            *scene;
	SurfelCollection *surfelsA,
		             *surfelsB,
					 *surfelsCSG;
	SurfelInterface  *surfel,
		             *newSurfel;
	Vector3D         newPosition,
					 newNormal,
					 newAxisOne,
					 newAxisTwo;

	bool		     insideA,
		             insideB;

	unsigned int	 nOfCSGSurfels;

	Transformer      *transformer;       // the scene transformer
	MyDataTypes::TransformationMatrix16f objectMatrix;


	scene = Scene::getInstance();
	transformer = Transformer::getSceneTransformer();

	numTotalQueries = 0;
	numBoundarySurfels = 0;


	/////////////////////////////////////////////////////////
	// classify surfels according to inside and outside
	//

	object1 = objectA;
	object2 = objectB;
	clusterTree = clusterTreeA;
	kdTree = kdTreeB;
	surfelsVector = surfelsVectorB;
	clipFlag = SurfelInterface::CLIP_BOUNDARY_A;
	classifySurfels ();

	object1 = objectB;
	object2 = objectA;
	clusterTree = clusterTreeB;
	kdTree = kdTreeA;
	surfelsVector = surfelsVectorA;
	clipFlag = SurfelInterface::CLIP_BOUNDARY_B;
	classifySurfels ();
	
	switch (operation) {
	
	case OP_A_union_B:
		insideA = false;
		insideB = false;
		break;

	case OP_A_intersection_B:
		insideA = true;
		insideB = true;
		break;

	case OP_A_minus_B:
		insideA = false;
		insideB = true;
		break;

	case OP_B_minus_A:
		insideA = true;
		insideB = false;
		break;
	}


	clipOrientation = insideA | insideB;


	surfelsA = objectA->getSurfelCollection();
	surfelsB = objectB->getSurfelCollection();

	// count CSG surfels
	nOfCSGSurfels = 0;
	for (surfel = surfelsA->getFirstSurfel(); surfel != 0; surfel = surfelsA->getNextSurfel()) {
		if (surfel->isFlagOn (SurfelInterface::INSIDE) == insideA) {
			nOfCSGSurfels++;
		}
	}
	for (surfel = surfelsB->getFirstSurfel(); surfel != 0; surfel = surfelsB->getNextSurfel()) {
		if (surfel->isFlagOn (SurfelInterface::INSIDE) == insideB) {
			nOfCSGSurfels++;
		}
	}

	// build csg surfel collection
	surfelsCSG = objectCSG->getSurfelCollection();
	surfelsCSG->clear(false);
	surfelsCSG->reserve(nOfCSGSurfels);	// reserve memory for CSG surfels

	// insert new surfels
	// transform position, normals and axes into scene coordinates
	// flip local frame if necessary

	objectA->getTransformationMatrix (objectMatrix);
	transformer->setObjectTransformationMatrix (objectMatrix);	

	for (surfel = surfelsA->getFirstSurfel(); surfel != 0; surfel = surfelsA->getNextSurfel()) {
		if (surfel->isFlagOn (SurfelInterface::INSIDE) == insideA) {
			newSurfel = surfelsCSG->addSurfel (surfel, false);
			newPosition = transformer->objectToScene (newSurfel->getPosition());
			scene->setSurfelPosition (newSurfel, newPosition, false);

			newNormal = transformer->objectToSceneVector (newSurfel->getNormal());
			newAxisOne = transformer->objectToSceneVector (newSurfel->getAxisOne());
			newAxisTwo = transformer->objectToSceneVector (newSurfel->getAxisTwo());

			newNormal.normalize();
			if (operation == OP_B_minus_A) {

				scene->setSurfelNormalTangentAxesSystem (	newSurfel, 
															-newNormal, 
															newAxisTwo,
															newAxisOne, 
															false);
			}
			else {
				scene->setSurfelNormalTangentAxesSystem (	newSurfel, 
															newNormal, 
															newAxisOne, 
															newAxisTwo, 
															false);
			}
		}
	}


	objectB->getTransformationMatrix (objectMatrix);
	transformer->setObjectTransformationMatrix (objectMatrix);
	
	for (surfel = surfelsB->getFirstSurfel(); surfel != 0; surfel = surfelsB->getNextSurfel()) {
		if (surfel->isFlagOn (SurfelInterface::INSIDE) == insideB) {
			newSurfel = surfelsCSG->addSurfel (surfel, false);
			newPosition = transformer->objectToScene (newSurfel->getPosition());
			scene->setSurfelPosition (newSurfel, newPosition, false);

			newNormal = transformer->objectToSceneVector (newSurfel->getNormal());
			newAxisOne = transformer->objectToSceneVector (newSurfel->getAxisOne());
			newAxisTwo = transformer->objectToSceneVector (newSurfel->getAxisTwo());

			newNormal.normalize();

			if (operation == OP_A_minus_B) {

				scene->setSurfelNormalTangentAxesSystem (	newSurfel, 
															-newNormal, 
															newAxisTwo,
															newAxisOne, 
															false);
			}
			else {
				scene->setSurfelNormalTangentAxesSystem (	newSurfel, 
															newNormal, 
															newAxisOne, 
															newAxisTwo, 
															false);
			}
			
		}
	}


	this->configurationWidget->totalTextLabel->setNum ((int) surfelsCSG->getNofSurfels());
	this->configurationWidget->boundaryTextLabel->setNum (numBoundarySurfels);
	this->configurationWidget->queryTextLabel->setNum (numTotalQueries);

}



void CSG_Tool::classifySurfels () {


	MyDataTypes::TransformationMatrix16f objectMatrix1,
		                                 objectMatrix2,
										 inverseMatrix2;
	Scene            *scene;

	ClusterNode      *node;

	scene = Scene::getInstance();


	// compute the transformation matrix that transforms from object1 to object2

	object1->getTransformationMatrix (objectMatrix1);
	object2->getTransformationMatrix (objectMatrix2);

	MtrInverse4x4f (objectMatrix2, inverseMatrix2);
	MtrMult4x4f (inverseMatrix2, objectMatrix1, objectToObjectTransformation);

	objectToObjectScaleFactor = MtrGetUniformScalingFactor4x4f (objectMatrix1) * MtrGetUniformScalingFactor4x4f (inverseMatrix2);


	node = clusterTree->getRoot();
	classifyNode (node);
}




////////////////////////////////////////////////////////////////////////////////////
// classifies all surfels by recursively traversing the sphere cluster tree.
// Only surfels whose center is inside are classified as such. Overlapping
// surfels are discarded.
// all surfels in a leaf are marked as CLIP_BOUNDARY_A/B. These are candidates for
// sampling the intersection curve
//


void CSG_Tool::classifyNode (ClusterNode *node) {

	Vector3D        center,
		            toPoint;
	float           squaredDistance;
	SurfelInterface **nodeSurfels;
	int				nofNodeSurfels;
	int             i;
	bool            inside;
	Scene           *scene;

	scene = Scene::getInstance();

	// transform to other objects local coordinates system;
	MtrMultVector3DHom (objectToObjectTransformation, node->getCenter(), &center);
	

	kdTree->queryPosition (center);
	numTotalQueries++;

	squaredDistance = kdTree->getSquaredDistance (0);


	if (node->getRadius() * node->getRadius() * objectToObjectScaleFactor * objectToObjectScaleFactor < squaredDistance) {

		toPoint = kdTree->getNeighbourPosition(0) - center;
		inside = ((Vector3D::dotProduct (toPoint, surfelsVector->at(kdTree->getNeighbourPositionIndex(0))->getNormal())) > 0);

		nofNodeSurfels = node->getNOfSurfels();
		nodeSurfels = node->getArray();

		for (i = 0; i < nofNodeSurfels; i++) {
			scene->setSurfelFlagOn (nodeSurfels[i], SurfelInterface::INSIDE, inside, false);
			
			scene->setSurfelFlagOn (nodeSurfels[i], clipFlag, false, false);
		}
	}

	else {

		if (node->isLeaf()) {

			nofNodeSurfels = node->getNOfSurfels();
			nodeSurfels = node->getArray();
			inside = false;
			for (i = 0; i < nofNodeSurfels; i++) {
		
				// transform to other objects local coordinates system;				
				MtrMultVector3DHom (objectToObjectTransformation, nodeSurfels[i]->getPosition(), &center);
				
				kdTree->queryPosition (center);
				numTotalQueries++;			
						
				toPoint = kdTree->getNeighbourPosition(0) - center;
				inside = ((Vector3D::dotProduct (toPoint, surfelsVector->at(kdTree->getNeighbourPositionIndex(0))->getNormal())) > 0);	

				scene->setSurfelFlagOn (nodeSurfels[i], SurfelInterface::INSIDE, inside, false);				

				numBoundarySurfels++;

				scene->setSurfelFlagOn (nodeSurfels[i], clipFlag, true, false);
			}
		}
		else {

			classifyNode (&node->getChildren()[0]);
			classifyNode (&node->getChildren()[1]);
		}
	}

}

void CSG_Tool::computeFeatureCurves (FeaturePoint* featurePoints, int nOfFeaturePoints) {

	Vector3D			*featurePositions;
	SurfelCollection	*surfelsCSG;
	KdTree				*featureKdTree;
	unsigned int		i, nOfFeatureSurfels;
	Scene				*scene;
	FeatureList		    *featureList;
	float				overlapFactor, minDistFactor, maxDistFactor, maxAngle;
	bool				isNum, bDoSubsample, bDoRefine, bDoRefineAngle;

	featureLists->clear();
	newFeaturesList->clear();

	if (nOfFeaturePoints < 4) {
		return;
	}

	nOfFeatureSurfels = 0;

	scene = Scene::getInstance();

	surfelsCSG = objectCSG->getSurfelCollection();
	featurePositions = new Vector3D[nOfFeaturePoints];

	for (i=0; i<nOfFeaturePoints; i++) {
		featurePositions[i] = featurePoints[i].position;
	}	

	featureKdTree = new KdTree (featurePositions, nOfFeaturePoints, 10);
	featureKdTree->setNOfNeighbours(20);

	minDistFactor = configurationWidget->MinDistanceLineEdit->text().toFloat(&isNum);
	if (!isNum || minDistFactor <= 0.0) {
		minDistFactor = 1.0;
		configurationWidget->MinDistanceLineEdit->setText("1.0");
	}
	maxDistFactor = configurationWidget->MaxDistanceLineEdit->text().toFloat(&isNum);
	if (!isNum || maxDistFactor <= 0.0) {
		maxDistFactor = 1.0;
		configurationWidget->MaxDistanceLineEdit->setText("1.0");
	}
	overlapFactor = configurationWidget->OverlapLineEdit->text().toFloat(&isNum);
	if (!isNum || overlapFactor <= 0.0) {
		overlapFactor = 1.0;
		configurationWidget->OverlapLineEdit->setText("1.0");
	}
	maxAngle = configurationWidget->MaxAngleSpinBox->value() * MY_PI / 180.0f;

	bDoSubsample	= configurationWidget->MinFeatureDistanceCheckBox->isChecked();
	bDoRefine		= configurationWidget->MaxFeatureDistanceCheckBox->isChecked();
	bDoRefineAngle	= configurationWidget->MaxAngleCheckBox->isChecked();

	for (i=0; i<nOfFeaturePoints; i++) {

		if (!featurePoints[i].isCovered) {
			featureList = computeCurve(&featurePoints[i], featurePoints, featureKdTree);

			if (featureList->count() >= 4) {
				featureLists->append(featureList);

				if (bDoSubsample) {
					subsampleCurve(featureList, minDistFactor);
				}
				if (bDoRefine) {
					refineCurve(surfelsCSG, featureList, maxDistFactor, newFeaturesList);
				}
				if (bDoRefineAngle) {
					refineAngle(surfelsCSG, featureList, maxAngle, newFeaturesList);
				}

				nOfFeatureSurfels += sampleCurve(surfelsCSG, featureList, overlapFactor);

			}
			else {
				delete featureList;
			}
		}
	}


	recomputeClipSurfels(surfelsCSG, nOfFeatureSurfels);

	delete	 featureKdTree;
	delete[] featurePositions;
}


bool CSG_Tool::adjustClipSurfel(SurfelInterface *clipSurfel, SurfelInterface *surfel, Scene *scene) {

	Vector3D	line, distance, clipSurfelPos, surfelPos;
	float		sqrDist1, sqrDist2, sqrDist, axisOneSqrLength, axisTwoSqrLength, lineLength, distanceSqrLength,
		        sqrB;

	clipSurfelPos	= clipSurfel->getPosition();
	surfelPos		= surfel->getPosition();
	line = clipSurfel->getAxisOne();
	lineLength = line.normalize();
	distance = surfelPos - clipSurfelPos;
	distanceSqrLength = distance.getSquaredLength();


	sqrB = Vector3D::dotProduct (line, distance);
	sqrB *= sqrB;

	sqrDist = distanceSqrLength - sqrB;

	if (sqrB > lineLength * lineLength) {
		// closest point is not on axis
		// take min distance to end-point of axis
		line *= lineLength;
		sqrDist1 = ((clipSurfelPos + line) - surfelPos).getSquaredLength();
		sqrDist2 = ((clipSurfelPos - line) - surfelPos).getSquaredLength();
		sqrDist = __min(sqrDist1, sqrDist2);
	}

	axisOneSqrLength = surfel->getAxisOne().getSquaredLength();
	axisTwoSqrLength = surfel->getAxisTwo().getSquaredLength();

	if (sqrDist >= __max(axisOneSqrLength, axisTwoSqrLength)) {
		return true;
	}
	
	if (sqrDist < __min(axisOneSqrLength, axisTwoSqrLength)*0.5*0.5) {
		// surfel is to small
		return false;
	}
	else {
		scene->setSurfelRadius(surfel, sqrt(sqrDist), false);
		return true;
	}
}

SurfelInterface* CSG_Tool::insertNewSurfel(SurfelCollection *surfelsCSG, Vector3D &pos, SurfelInterface *surfel1, SurfelInterface *surfel2, float fac, SurfelInterface::Flag flag) {

	SurfelInterface *newSurfel = surfelsCSG->addSurfel(surfel1, false);
	Scene			*scene = Scene::getInstance();
	Vector3D		newPosition, newAxisOne, newAxisTwo, newNormal, dist, axisOne, axisTwo;
	float			axisTwoLength;

	newPosition		= surfel1->getPosition() + fac * (surfel2->getPosition() - surfel1->getPosition());
	newNormal		= surfel1->getNormal() * fac + surfel2->getNormal() * (1.0 - fac);
	newNormal.normalize();
	newAxisTwo		= pos - newPosition;
	axisTwoLength	= newAxisTwo.normalize();
	newAxisOne		= Vector3D::crossProduct(newAxisTwo, newNormal);
	newAxisOne.normalize();
	newAxisTwo		= Vector3D::crossProduct(newNormal, newAxisOne);
	newAxisTwo	   *= axisTwoLength;	
	newAxisOne	   *= (surfel1->getRadius() + surfel2->getRadius()) * 0.5;
	
	scene->setSurfelPosition(newSurfel, newPosition, false);
	scene->setSurfelNormalTangentAxesSystem(newSurfel, newNormal, newAxisOne, newAxisTwo, false);

	scene->setSurfelDiffuseColor(newSurfel, mixColor(surfel1->getDiffuseColor(), surfel2->getDiffuseColor(), fac), false);
	scene->setSurfelSpecularColor(newSurfel, mixColor(surfel1->getSpecularColor(), surfel2->getSpecularColor(), fac), false);

	scene->setSurfelFlagOn(newSurfel, flag, true, false);
	scene->setSurfelFlagOn(newSurfel, SurfelInterface::FEATURE, false, false);
	scene->setSurfelFlagOn(newSurfel, SurfelInterface::INVISIBLE, false, false);
	scene->clearAssociatedSurfels(newSurfel, false);

	return newSurfel;
}


void CSG_Tool::recomputeClipSurfels(SurfelCollection *surfels, unsigned int nOfFeatureSurfels) {

	Scene			*scene;
	KdTree			*kdTree;
	Vector3D		*featurePositions;
	SurfelInterface	*surfel,
					**featureSurfels,
					*featureSurfel;
	unsigned int	i, nOfFoundNeighbors;
	bool			deleteSurfel;
	float			len1, len2;

	if (nOfFeatureSurfels == 0) {
		return;
	}
	
	scene = Scene::getInstance();

	featurePositions	= new Vector3D[nOfFeatureSurfels];
	featureSurfels		= new SurfelInterface*[nOfFeatureSurfels];

	for (i=0, surfel = surfels->getFirstSurfel(); surfel; surfel = surfels->getNextSurfel()) {

		if (surfel->isFlagOn(SurfelInterface::FEATURE)) {
			featurePositions[i] = surfel->getPosition();
			featureSurfels[i]	= surfel;
			i++;
		}

	}
	
	kdTree = new KdTree(featurePositions, nOfFeatureSurfels, 10);
	kdTree->setNOfNeighbours(4);

	for (surfel = surfels->getFirstSurfel(); surfel; ) {

		if ( surfel->isFlagOn(SurfelInterface::CLIP_BOUNDARY_A) ||
			 surfel->isFlagOn(SurfelInterface::CLIP_BOUNDARY_B)
			) {

			kdTree->queryPosition(surfel->getPosition());
			nOfFoundNeighbors = kdTree->getNOfFoundNeighbours();
			deleteSurfel = false;

			len1 = surfel->getAxisOne().getSquaredLength();
			len2 = surfel->getAxisTwo().getSquaredLength();

			if (kdTree->getSquaredDistance(0) < __max(len1, len2)) {
				// test if surfel is on the correct side of the clip-line
				if (surfel->isFlagOn(SurfelInterface::CLIP_BOUNDARY_A)) {
					featureSurfel = featureSurfels[kdTree->getNeighbourPositionIndex(0)];
				}
				else {
					featureSurfel = featureSurfels[kdTree->getNeighbourPositionIndex(0)]->getAssociatedSurfel(0);
				}
		
				if (Vector3D::dotProduct(featureSurfel->getAxisTwo(), surfel->getPosition() - featureSurfel->getPosition()) < 0) {
					scene->setSurfelFlagOn(surfel, SurfelInterface::INVISIBLE, true, false);
					surfel = surfels->getNextSurfel();
					continue;
				}
				
			}

			for (i=0; i<nOfFoundNeighbors; i++) {

				if (!adjustClipSurfel(featureSurfels[kdTree->getNeighbourPositionIndex(i)], surfel, scene)) {
					scene->setSurfelFlagOn(surfel, SurfelInterface::INVISIBLE, true, false);
					break;
				}
			}

			surfel = surfels->getNextSurfel();

		}
		else {
			surfel = surfels->getNextSurfel();
		}

	}

	delete		kdTree;
	delete[]	featurePositions;
	delete[]	featureSurfels;
}


void CSG_Tool::subsampleCurve(FeatureList *featureList, const float minDistFactor) {

	FeaturePoint		*prev, *current, *next;
	float				dist;
	float				prevRadius, currentRadius, nextRadius;
	float				radius1, radius2;

	prev = featureList->first();

	if (!prev) {
		return;
	}

	radius1 = prev->surfel1->getRadius();
	radius2 = prev->surfel2->getRadius();
	prevRadius = __min(radius1, radius2);

	while (current = featureList->next()) {
		radius1 = current->surfel1->getRadius();
		radius2 = current->surfel2->getRadius();
		currentRadius = __min(radius1, radius2);

		// delete current point if distance of prev and next is smaller than the sum of their radii
		while (next = featureList->getNext()) {

			radius1 = current->surfel2->getRadius();
			radius2 = next->surfel2->getRadius();
			nextRadius = __min(radius1, radius2);
			dist = (prev->position - next->position).getLength();
			if (dist < (prevRadius + nextRadius) * minDistFactor) {
				featureList->setAsDeleted();
				featureList->next();
				current = next;
				currentRadius = nextRadius;
			}
			else {
				break;
			}			
		}		

		prev		= current;
		prevRadius	= currentRadius;
	}

}

void CSG_Tool::refineCurve(SurfelCollection *surfelsCSG, FeatureList *featureList, const float maxDistFactor, FeatureList *newFeatureList) {

	FeaturePoint		*prev2, *prev, *current, *next;
	FeaturePoint		*newPoint;
	float				dist;
	float				prevRadius, currentRadius;
	Vector3D			prevDist, nextDist;
	const float			cWeight = 0.125;
	float				len1, len2;
	float				fac;
	
	prev = featureList->first();
	if (!prev) {
		return;
	}

	prev2 = NULL;
	len1 = prev->surfel1->getAxisOne().getLength();
	len2 = prev->surfel2->getAxisOne().getLength();
	prevRadius = __min(len1, len2);
	current = featureList->next();

	while (current) {

		dist = (prev->position - current->position).getLength();
		len1 = current->surfel1->getAxisOne().getLength();
		len2 = current->surfel2->getAxisOne().getLength();
		currentRadius = __min(len1, len2);

		if (dist > (prevRadius + currentRadius) * maxDistFactor) {
			// insert a point if distance of prev and current is too large

			newPoint = new FeaturePoint;
			newFeatureList->append(newPoint);
			newPoint->isCovered = false;
			next = featureList->getNext();
			
			if (!prev2 || !next) {
				newPoint->position = (prev->position + current->position) * 0.5;
				fac = 0.5;
			}
			else {
				newPoint->position = - cWeight*prev2->position
									 + (cWeight+0.5)*prev->position 
									 + (cWeight+0.5)*current->position
									 - cWeight*next->position;

				fac = sqrt((newPoint->position - prev->position).getSquaredLength() / (current->position - prev->position).getSquaredLength());

				if (fac < 0.25 || fac > 0.75) {
					// bad insertion, choose middle
					newPoint->position = (prev->position + current->position) * 0.5;
					fac = 0.5;
				}
			}

			newPoint->surfel1 = insertNewSurfel(surfelsCSG, newPoint->position, prev->surfel1, current->surfel1, fac, SurfelInterface::CLIP_BOUNDARY_A);
			newPoint->surfel2 = insertNewSurfel(surfelsCSG, newPoint->position, prev->surfel2, current->surfel2, fac, SurfelInterface::CLIP_BOUNDARY_B);
		
			featureList->prev();
			featureList->insert(newPoint);
			current = newPoint;
		}
		else {
			prev2		= prev;
			prev		= current;
			prevRadius	= currentRadius;
			current		= featureList->next();
		}
	}

}

void CSG_Tool::refineCurve(SurfelCollection *surfelsCSG, FeatureList *featureList, FeatureList *newFeatureList) {

	FeaturePoint		*prev2, *prev, *current, *next;
	FeaturePoint		*newPoint;
	const float			cWeight = 0.125;
	float				fac;
	
	prev = featureList->first();
	prev2 = NULL;
	current = featureList->next();
	
	while (current) {
		
		newPoint = new FeaturePoint;
		newFeatureList->append(newPoint);
		newPoint->isCovered = false;
		next = featureList->getNext();
			
		if (!prev2 || !next) {
			newPoint->position = (prev->position + current->position) * 0.5;
			fac = 0.5;
		}
		else {
			newPoint->position = - cWeight*prev2->position
								 + (cWeight+0.5)*prev->position 
								 + (cWeight+0.5)*current->position
								 - cWeight*next->position;
			fac = sqrt((newPoint->position - prev->position).getSquaredLength() / (current->position - prev->position).getSquaredLength());

			if (fac < 0.25 || fac > 0.75) {
				// bad insertion, choose middle
				newPoint->position = (prev->position + current->position) * 0.5;
				fac = 0.5;
			}

		}

		newPoint->surfel1 = insertNewSurfel(surfelsCSG, newPoint->position, prev->surfel1, current->surfel1, fac, SurfelInterface::CLIP_BOUNDARY_A);
		newPoint->surfel2 = insertNewSurfel(surfelsCSG, newPoint->position, prev->surfel2, current->surfel2, fac, SurfelInterface::CLIP_BOUNDARY_B);

		featureList->prev();
		featureList->insert(newPoint);	// insert new point between prev and current
		featureList->next();
		
		prev2	= prev;
		prev	= current;
		current = next;
		next	= featureList->next();
	}
}


void CSG_Tool::refineAngle(SurfelCollection *surfelsCSG, FeatureList *featureList, const float maxAngle, FeatureList *newFeatureList) {

	FeaturePoint		*prev2, *prev, *current, *next, *next2;
	FeaturePoint		*newPoint;
	Vector3D			prevDist, nextDist, newPosition;
	float				angle;
	const float			cWeight = 0.125;
	const float			cMinLength  = (float)1.0e-7;
	const float			cMinSqrLength = cMinLength * cMinLength;
	float				prevDistLength, nextDistLength, fac;
	bool				newPointsInserted;
	
	for (int i=0; i<2; i++) {
		newPointsInserted = false;
		prev2 = featureList->first();
		prev = featureList->next();
		current = featureList->next();

		if (current) {
			prevDist = current->position - prev->position;
			prevDistLength = prevDist.normalize();
		}

		while (next=featureList->next()) {

			
			
			nextDist = next->position - current->position;
			nextDistLength = nextDist.normalize();

			if (prevDistLength < cMinLength || nextDistLength < cMinLength) {
				prev2			= prev;
				prev			= current;
				current			= next;
				prevDist		= nextDist;
				prevDistLength	= nextDistLength;
				continue;
			}

			angle = fabs(Vector3D::dotProduct(prevDist, nextDist));
			if (angle > 1.0) {
				angle = 1.0;
			}
			angle = acos(angle);

			if (angle > maxAngle) {

				next2 = featureList->getNext();

				newPosition =	- cWeight*prev2->position
					 			+ (cWeight+0.5)*prev->position 
					 			+ (cWeight+0.5)*current->position
								- cWeight*next->position;

				if ( (newPosition - current->position).getSquaredLength() > cMinSqrLength) {

					newPoint = new FeaturePoint;
					newFeatureList->append(newPoint);
					newPoint->isCovered = false;
					newPoint->position  = newPosition;

					fac = (newPoint->position - prev->position).getLength() / prevDistLength;
					newPoint->surfel1 = insertNewSurfel(surfelsCSG, newPoint->position, prev->surfel1, current->surfel1, fac, SurfelInterface::CLIP_BOUNDARY_A);
					newPoint->surfel2 = insertNewSurfel(surfelsCSG, newPoint->position, prev->surfel2, current->surfel2, fac, SurfelInterface::CLIP_BOUNDARY_B);

					featureList->prev();		// set featureList to current
					featureList->prev();		// set featureList to prev
					featureList->insert(newPoint);	// insert between prev and current
					featureList->next();		// set featureList to current
					featureList->next();		// set featureList to next
					newPointsInserted = true;
				}

				if (next2) {
					newPosition =	- cWeight*prev->position
					 				+ (cWeight+0.5)*current->position 
					 				+ (cWeight+0.5)*next->position
									- cWeight*next2->position;

					if ( (newPosition - current->position).getSquaredLength() > cMinSqrLength) {
						newPoint = new FeaturePoint;
						newFeatureList->append(newPoint);
						newPoint->isCovered = false;
						newPoint->position	= newPosition;

						fac = (newPoint->position - current->position).getLength() / nextDistLength;
						newPoint->surfel1 = insertNewSurfel(surfelsCSG, newPoint->position, current->surfel1, next->surfel1, fac, SurfelInterface::CLIP_BOUNDARY_A);
						newPoint->surfel2 = insertNewSurfel(surfelsCSG, newPoint->position, current->surfel2, next->surfel2, fac, SurfelInterface::CLIP_BOUNDARY_B);

						featureList->prev();			// set featureList to current
						featureList->insert(newPoint);	// insert between current and next
						featureList->next();			// set featureList to next
						newPointsInserted = true;
					}

					prev2 = current;
					prev = newPoint;
					current = next;
					prevDist = current->position - prev->position;
					prevDistLength = prevDist.normalize();

					continue;
				}
				else {
					break;
				}
			}

			prev2			= prev;
			prev			= current;
			current			= next;
			prevDist		= nextDist;
			prevDistLength	= nextDistLength;
		}

		if (!newPointsInserted) {
			break;
		}

	}


}


unsigned int CSG_Tool::sampleCurve(SurfelCollection *surfelsCSG, FeatureList *featureList, float overlappingFactor) {

	FeaturePoint	*prev, *next;
	Vector3D		distance, midPoint;
	Vector3D		newNormal1, newNormal2, newPosition, newAxisOne1, newAxisOne2, newAxisTwo1, newAxisTwo2;
	SurfelInterface *newSurfel1, *newSurfel2;
	SurfelInterface *prevSurfel1, *prevSurfel2, *nextSurfel1, *nextSurfel2;
	float			length, prevDist1, prevDist2, nextDist1, nextDist2, distFac1, distFac2;
	Scene			*scene;
	QColor			prevColor, nextColor;
	float			radius;
	SurfelInterface	*prevNewSurfel1, *prevNewSurfel2, *newSurfel;
	Vector3D		prevAxisOne;
	const float		cCosNormals = cos(MY_PI/4.0);
	unsigned int	nOfFeatureSurfels = 0;

	prevNewSurfel1 = prevNewSurfel2 = NULL;
	scene = Scene::getInstance();
	prev = featureList->first();

	if (featureList->getFirst() == featureList->getLast()) {
		// the feature list builds a circle
		// we append the second surfel. In this way, one clip surfel is added twice, but therefore we get
		// also the surfels on the bisection line
		featureList->append(featureList->getNext());
		featureList->first();
	}


	while (next = featureList->next()) {

		prevSurfel1	= prev->surfel1;
		prevSurfel2	= prev->surfel2;
		nextSurfel1	= next->surfel1;
		nextSurfel2 = next->surfel2;
	
		distance = next->position - prev->position;

		if (distance.getSquaredLength() > 0.00001f) {
		
			newPosition = (prev->position + next->position) * 0.5;			
			newAxisOne1	= distance * (0.5 * overlappingFactor);
			length		= newAxisOne1.getLength();
			
			prevDist1	= (newPosition - prevSurfel1->getPosition()).getLength();
			prevDist2	= (newPosition - prevSurfel2->getPosition()).getLength();
			nextDist1	= (newPosition - nextSurfel1->getPosition()).getLength();
			nextDist2	= (newPosition - nextSurfel2->getPosition()).getLength();
			distFac1	= prevDist1 / (prevDist1 + nextDist1);
			distFac2	= prevDist2 / (prevDist2 + nextDist2);		

			midPoint	= (prevSurfel1->getPosition() + nextSurfel1->getPosition()) * 0.5;
			newNormal1 = distFac1 * prevSurfel1->getNormal() + ((1.0 - distFac1) * nextSurfel1->getNormal());
			newAxisTwo1 = Vector3D::crossProduct(newNormal1, newAxisOne1);
			newAxisTwo1.normalize();
			newAxisTwo1 *= length;
			newNormal1   = Vector3D::crossProduct(newAxisOne1, newAxisTwo1);
			newNormal1.normalize();

			midPoint	= (prevSurfel2->getPosition() + nextSurfel2->getPosition()) * 0.5;
			newNormal2 = distFac2 * prevSurfel2->getNormal() + ((1.0 - distFac2) * nextSurfel2->getNormal());
			newAxisOne2 = newAxisOne1;
			newAxisTwo2 = Vector3D::crossProduct(newNormal2, newAxisOne1);
			newAxisTwo2.normalize();
			newAxisTwo2 *= length;
			newNormal2   = Vector3D::crossProduct(newAxisOne2, newAxisTwo2);
			newNormal2.normalize();

			int flip = clipOrientation ? -1 : 1;
			if (Vector3D::dotProduct(newAxisTwo1, newNormal2)*flip < 0) {
				newAxisTwo1 = -newAxisTwo1;
				newAxisOne1 = -newAxisOne1;
			}
			if (Vector3D::dotProduct(newAxisTwo2, newNormal1)*flip < 0) {
				newAxisTwo2 = -newAxisTwo2;
				newAxisOne2 = -newAxisOne2;
			}

			
			newSurfel1	= surfelsCSG->addSurfel(prevSurfel1, false);
			newSurfel2	= surfelsCSG->addSurfel(prevSurfel2, false);
			
			scene->setSurfelPosition(newSurfel1, newPosition, false);
			scene->setSurfelPosition(newSurfel2, newPosition, false);
			scene->setSurfelNormalTangentAxesSystem(newSurfel1, newNormal1, newAxisOne1, newAxisTwo1, false);
			scene->setSurfelNormalTangentAxesSystem(newSurfel2, newNormal2, newAxisOne2, newAxisTwo2, false);
			
			scene->setSurfelDiffuseColor(newSurfel1, mixColor(prevSurfel1->getDiffuseColor(), nextSurfel1->getDiffuseColor(), distFac1), false);
			scene->setSurfelDiffuseColor(newSurfel2, mixColor(prevSurfel2->getDiffuseColor(), nextSurfel2->getDiffuseColor(), distFac2), false);
			scene->setSurfelSpecularColor(newSurfel1, mixColor(prevSurfel1->getSpecularColor(), nextSurfel1->getSpecularColor(), distFac1), false);
			scene->setSurfelSpecularColor(newSurfel2, mixColor(prevSurfel2->getSpecularColor(), nextSurfel2->getSpecularColor(), distFac2), false);
			scene->clearAssociatedSurfels(newSurfel1, false);
			scene->clearAssociatedSurfels(newSurfel2, false);
			scene->addAssociatedSurfel(newSurfel1, newSurfel2, false);
			scene->addAssociatedSurfel(newSurfel2, newSurfel1, false);
			scene->setSurfelFlagOn(newSurfel1, SurfelInterface::CLIP_ORIENTATION, clipOrientation, false);
			scene->setSurfelFlagOn(newSurfel2, SurfelInterface::CLIP_ORIENTATION, clipOrientation, false);
			scene->setSurfelFlagOn(newSurfel1, SurfelInterface::FEATURE, true, false);
			scene->setSurfelFlagOn(newSurfel2, SurfelInterface::FEATURE, false, false);
			scene->setSurfelFlagOn(newSurfel1, SurfelInterface::CLIP_BOUNDARY_A, false, false);
			scene->setSurfelFlagOn(newSurfel2, SurfelInterface::CLIP_BOUNDARY_B, false, false);
			scene->setSurfelFlagOn(newSurfel1, SurfelInterface::INVISIBLE, false, false);
			scene->setSurfelFlagOn(newSurfel2, SurfelInterface::INVISIBLE, false, false);

			nOfFeatureSurfels++;

			newAxisOne1.normalize();
			if (prevNewSurfel1) {
				// insert a new surfel between two clipped surfels
				
				// compute the bisecting line of both axis-ones of the adjacent surfels
				newAxisTwo1 = prevNewSurfel1->getAxisTwo();
				length = newAxisTwo1.normalize();
				newAxisTwo2 = newSurfel1->getAxisTwo();
				length += newAxisTwo2.normalize();
				length /= 2.0;
				newAxisTwo1 += newAxisTwo2;
				newAxisTwo1.normalize();
				newPosition = prev->position + newAxisTwo1*length;
				radius		= (Vector3D::crossProduct(prevAxisOne, newPosition - prev->position)).getLength();

				if (radius > 0.0f) {

					newSurfel	= surfelsCSG->addSurfel(prevNewSurfel1, false);
					computeSurfelProperties(newSurfel, newPosition, radius, clipKdTreeA, 4, clipSurfelsA, scene);
					scene->clearAssociatedSurfels(newSurfel, false);
					scene->setSurfelFlagOn(newSurfel, SurfelInterface::FEATURE, false, false);
					scene->setSurfelFlagOn(newSurfel, SurfelInterface::CLIP_BOUNDARY_A, true, false);
					scene->setSurfelFlagOn(newSurfel, SurfelInterface::INVISIBLE, false, false);
			
					newAxisTwo1 = prevNewSurfel2->getAxisTwo();
					length = newAxisTwo1.normalize();
					newAxisTwo2 = newSurfel2->getAxisTwo();
					length += newAxisTwo2.normalize();
					length /= 2.0;
					newAxisTwo1 += newAxisTwo2;
					newAxisTwo1.normalize();
					newPosition = prev->position + newAxisTwo1*length;
					radius		= (Vector3D::crossProduct(prevAxisOne, newPosition - prev->position)).getLength();

					newSurfel	= surfelsCSG->addSurfel(prevNewSurfel2, false);
					computeSurfelProperties(newSurfel, newPosition, radius, clipKdTreeB, 4, clipSurfelsB, scene);
					scene->clearAssociatedSurfels(newSurfel, false);
					scene->setSurfelFlagOn(newSurfel, SurfelInterface::FEATURE, false, false);
					scene->setSurfelFlagOn(newSurfel, SurfelInterface::CLIP_BOUNDARY_B, true, false);
					scene->setSurfelFlagOn(newSurfel, SurfelInterface::INVISIBLE, false, false);
				}
			} // if (prevNewSurfel1)

			prevNewSurfel1 = newSurfel1;
			prevNewSurfel2 = newSurfel2;
			prevAxisOne	   = newAxisOne1;
		
		}

		prev = next;

	} // if (distance.getSquaredLength() > 0.0f)

	return nOfFeatureSurfels;
}


void CSG_Tool::computeSurfelProperties(SurfelInterface *newSurfel, Vector3D &newPosition, float radius, KdTree *kdTree, unsigned int nOfNeighbors, SurfelInterface **surfels, Scene *scene) {
	unsigned int	nOfFoundNeighbors, i;
	float			maxSqrDist, sqrSigma, weight, sumWeight;
	Vector3D		normal;
	QColor			newColor;
	int				diffR, diffG, diffB, specR, specG, specB;
	SurfelInterface *surfel;

	diffR = diffG = diffB = specR = specG = specB = 0.0f;
	normal = Vector3D(0,0,0);
	
	kdTree->setNOfNeighbours(nOfNeighbors);
	kdTree->queryPosition(newPosition);
	nOfFoundNeighbors = kdTree->getNOfFoundNeighbours();
	
	ASSERT(nOfFoundNeighbors);

	maxSqrDist = kdTree->getSquaredDistance(0);
	sqrSigma = -maxSqrDist / log(0.01);
	sumWeight = 0.0f;
	
	for (i=0; i<nOfFoundNeighbors; i++) {
		surfel = surfels[kdTree->getNeighbourPositionIndex(i)];
		weight = exp(maxSqrDist / sqrSigma);
		sumWeight += weight;
		
		normal += weight * surfel->getNormal();
		newColor = QColor(surfel->getDiffuseColor());
		diffR += weight*newColor.red();
		diffG += weight*newColor.green();
		diffB += weight*newColor.blue();
		newColor = QColor(surfel->getSpecularColor());
		specR += weight*newColor.red();
		specG += weight*newColor.green();
		specB += weight*newColor.blue();	
	}

	normal.normalize();
	sumWeight = 1.0f / sumWeight;
	diffR *= sumWeight;
	diffG *= sumWeight;
	diffB *= sumWeight;
	specR *= sumWeight;
	specG *= sumWeight;
	specB *= sumWeight;

	scene->setSurfelPosition(newSurfel, newPosition, false);
	scene->setSurfelNormal(newSurfel, normal, false);
	scene->setSurfelRadius(newSurfel, radius, false);
	scene->setSurfelDiffuseColor(newSurfel, QColor(diffR, diffG, diffB).rgb(), false);
	scene->setSurfelSpecularColor(newSurfel, QColor(specR, specG, specB).rgb(), false);
}


FeatureList* CSG_Tool::computeCurve(FeaturePoint* startPoint, FeaturePoint *featurePoints, KdTree *featureKdTree) {

	FeaturePoint	*nextPoint,
					*prevPoint,
					*currentPoint;
	FeatureList		*list;
	Scene			*scene;
	Vector3D		prevDistance,
					nextDistance,
					prev2Distance;
	unsigned int	nOfFoundNeighbors, i, prevI;

	scene = Scene::getInstance();

	list = new FeatureList();

	featureKdTree->queryPosition(startPoint->position);	

	// if there is already a covered feature point (a point belonging to a feature curve) nearby this point return
	for (i=1; i<featureKdTree->getNOfFoundNeighbours(); i++) {
		if (featurePoints[featureKdTree->getNeighbourPositionIndex(i)].isCovered) {
			return list;
		}
	}	

	list->append(startPoint);
	startPoint->isCovered = true;

	for (i=1; i<featureKdTree->getNOfFoundNeighbours(); i++) {
		if (featureKdTree->getSquaredDistance(i) > 0.0) {
			break;
		}
	}	
	nextPoint = &featurePoints[featureKdTree->getNeighbourPositionIndex(i)];

	list->append(nextPoint);
	nextPoint->isCovered = true;

	nextDistance =  nextPoint->position - startPoint->position;
	currentPoint = startPoint;
	prevDistance = Vector3D(0,0,0);
	i = 0;

	// search for neighbor surfels in one direction
	for (;;) {
		do {
			prevPoint	 = currentPoint;
			currentPoint = nextPoint;		

			prev2Distance = prevDistance;
			prevDistance = -nextDistance;
			featureKdTree->queryPosition(currentPoint->position);	
			prevI		= i;
			nOfFoundNeighbors = featureKdTree->getNOfFoundNeighbours();
			for (i=1; i<nOfFoundNeighbors; i++) {
				nextPoint = &featurePoints[featureKdTree->getNeighbourPositionIndex(i)];
				nextDistance = nextPoint->position - currentPoint->position;
				if ( Vector3D::dotProduct(nextDistance, prevDistance) <= 0) {
					if (nextPoint->isCovered) {
						if (nextPoint == list->getFirst()) {
							list->prepend(list->getLast());
							return list;
						}
					}
					else {
						list->append(nextPoint);
						nextPoint->isCovered = true;
						break;
					}				
				}
			}

		} while( i < nOfFoundNeighbors);

		if (prevI > 0) {
			// the previous search may lead to a dead end. Therefore we backtrap one step
			// and try again if there is another point. If so, we remove the last point again from the
			// list and continue with the new point.
			prevDistance = prev2Distance;
			currentPoint = prevPoint;

			featureKdTree->queryPosition(currentPoint->position);	
			for (i=prevI+1; i<featureKdTree->getNOfFoundNeighbours(); i++) {
				nextPoint = &featurePoints[featureKdTree->getNeighbourPositionIndex(i)];
				nextDistance = nextPoint->position - currentPoint->position;
				if ( Vector3D::dotProduct(nextDistance, prevDistance) <= 0) {
					if (nextPoint->isCovered) {
						if (nextPoint == list->getFirst()) {
							list->current()->isCovered = false;
							list->remove();
							list->prepend(list->getLast());
							return list;
						}
					}
					else {
						list->current()->isCovered = false;
						list->remove();
						list->append(nextPoint);
						nextPoint->isCovered = true;
						break;
					}				
				}
			}
			if (i == featureKdTree->getNOfFoundNeighbours()) {
				break;
			}
		}
		else {
			break;
		}

		
	}

	nextPoint = list->first();
	currentPoint = list->next();
	i = 0;	
	nextDistance =  nextPoint->position - currentPoint->position;

	// search for neighbor surfels in other direction
	for (;;) {
		do {
			prevPoint	 = currentPoint;
			currentPoint = nextPoint;		

			prev2Distance = prevDistance;
			prevDistance = -nextDistance;
			featureKdTree->queryPosition(currentPoint->position);	
			prevI		= i;

			for (i=1; i<featureKdTree->getNOfFoundNeighbours(); i++) {
								
				nextPoint = &featurePoints[featureKdTree->getNeighbourPositionIndex(i)];		
				nextDistance = nextPoint->position - currentPoint->position;
				if ( Vector3D::dotProduct(nextDistance, prevDistance) <= 0) {
					if (nextPoint->isCovered) {
						if (nextPoint == list->getLast()) {
							list->append(list->getFirst());
							return list;
						}
					}
					else {
						list->prepend(nextPoint);
						nextPoint->isCovered = true;
						break;
					}				
				}
			}

		} while( i < featureKdTree->getNOfFoundNeighbours());

		if (prevI > 0) {
			// the previous search may lead to a dead end. Therefore we backtrap one step
			// and try again if there is another point. If so, we remove the last point again from the
			// list and continue with the new point.			prevDistance = prev2Distance;
			currentPoint = prevPoint;

			featureKdTree->queryPosition(currentPoint->position);	
			for (i=prevI+1; i<featureKdTree->getNOfFoundNeighbours(); i++) {
				nextPoint = &featurePoints[featureKdTree->getNeighbourPositionIndex(i)];
				nextDistance = nextPoint->position - currentPoint->position;
				if ( Vector3D::dotProduct(nextDistance, prevDistance) <= 0) {
					if (nextPoint->isCovered) {
						if (nextPoint == list->getFirst()) {
							list->current()->isCovered = false;
							list->remove();
							list->append(list->getLast());
							return list;
						}
					}
					else {
						list->current()->isCovered = false;
						list->remove();
						list->prepend(nextPoint);
						nextPoint->isCovered = true;
						break;
					}				
				}
			}
			if (i == featureKdTree->getNOfFoundNeighbours()) {
				break;
			}
		}
		else {
			break;
		}
	
	}

	return list;	
}


void CSG_Tool::sampleIntersectionCurves () {
	

	SurfelInterface  *surfel,
		             *closestSurfel;
	SurfelCollection *surfelsCSG;
	Vector3D         center,
					 newPosition,
					 newAxisOne,
					 newAxisTwo1,
					 newAxisTwo2;
	Scene            *scene;
	float            distance;


	int               nofClipSurfelsA, 
		              nofClipSurfelsB,
					  i,
					  clipIndexA,
					  clipIndexB,
					  nOfFeatureSurfels;
	


	scene = Scene::getInstance();

	surfelsCSG = objectCSG->getSurfelCollection();

	if (clipPositionsA) {
		delete[] clipPositionsA;
		delete[] clipPositionsB;
		delete[] clipSurfelsA;
		delete[] clipSurfelsB;
		delete[] clipSurfelsOtherA;
		delete[] clipSurfelsOtherB;
		delete   clipKdTreeA;
		delete   clipKdTreeB;
		clipPositionsA		= NULL;
		clipPositionsB		= NULL;
		clipSurfelsA		= NULL;
		clipSurfelsB		= NULL;
		clipSurfelsOtherA	= NULL;
		clipSurfelsOtherB	= NULL;
		clipKdTreeA			= NULL;
		clipKdTreeB			= NULL;
	}
		

	////////////////////////////////////////////////////////////////
	// create data structures for all clip candidate pairs

	nofClipSurfelsA = 0;
	nofClipSurfelsB = 0;
	nOfFeatureSurfels = 0;

	for (surfel = surfelsCSG->getFirstSurfel(); surfel != 0; surfel = surfelsCSG->getNextSurfel()) {
		if (surfel->isFlagOn (SurfelInterface::CLIP_BOUNDARY_A)) {
			nofClipSurfelsA++;
		}
		else if (surfel->isFlagOn (SurfelInterface::CLIP_BOUNDARY_B)) {
			nofClipSurfelsB++;
		}
		scene->setSurfelFlagOn(surfel, SurfelInterface::FEATURE, false, false);
		scene->setSurfelFlagOn(surfel, SurfelInterface::COVERED, false, false);
	}

	if (nofClipSurfelsA*nofClipSurfelsB == 0) {
		return;
	}

	clipPositionsA = new Vector3D [nofClipSurfelsA];
	clipPositionsB = new Vector3D [nofClipSurfelsB];
	clipSurfelsA = new SurfelInterface* [nofClipSurfelsA];
	clipSurfelsB = new SurfelInterface* [nofClipSurfelsB];
	clipSurfelsOtherA = new int [nofClipSurfelsA];
	clipSurfelsOtherB = new int [nofClipSurfelsB];

	clipIndexA = clipIndexB = 0;

	for (surfel = surfelsCSG->getFirstSurfel(); surfel != 0; surfel = surfelsCSG->getNextSurfel()) {
		if (surfel->isFlagOn (SurfelInterface::CLIP_BOUNDARY_A)) {
			clipPositionsA[clipIndexA] = surfel->getPosition();
			clipSurfelsA[clipIndexA] = surfel;
			clipIndexA++;
		}
		else if (surfel->isFlagOn (SurfelInterface::CLIP_BOUNDARY_B)) {
			clipPositionsB[clipIndexB] = surfel->getPosition();
			clipSurfelsB[clipIndexB] = surfel;
			clipIndexB++;
		}
	}

	clipKdTreeA = new KdTree (clipPositionsA, nofClipSurfelsA, 10);
	clipKdTreeB = new KdTree (clipPositionsB, nofClipSurfelsB, 10);

	clipKdTreeA->setNOfNeighbours (1);
	clipKdTreeB->setNOfNeighbours (1);


	////////////////////////////////////////////////////////////////
	// determine if surfel should be feature surfel and find
	// corresponding clip surfel

	for (i = 0; i < nofClipSurfelsA; i++) {

		clipKdTreeB->queryPosition (clipPositionsA[i]);		
		closestSurfel = clipSurfelsB [clipKdTreeB->getNeighbourPositionIndex(0)];
			
		newPosition = this->sampleFeatureSurfelPosition (clipSurfelsA[i]->getPosition(),
			closestSurfel->getPosition(), clipSurfelsA[i]->getNormal(), closestSurfel->getNormal());

		distance = (newPosition - clipPositionsA[i]).getLength();
		if (distance < clipSurfelsA[i]->getRadius()) {
			clipSurfelsOtherA [i] = clipKdTreeB->getNeighbourPositionIndex(0);
		}
		else {
			clipSurfelsOtherA [i] = -1;
		}
	}

	for (i = 0; i < nofClipSurfelsB; i++) {

		clipKdTreeA->queryPosition (clipPositionsB[i]);		
		closestSurfel = clipSurfelsA [clipKdTreeA->getNeighbourPositionIndex(0)];
			
		newPosition = this->sampleFeatureSurfelPosition (clipSurfelsB[i]->getPosition(),
			closestSurfel->getPosition(), clipSurfelsB[i]->getNormal(), closestSurfel->getNormal());

		distance = (newPosition - clipPositionsB[i]).getLength();			
		if (distance < clipSurfelsB[i]->getRadius()) {
			clipSurfelsOtherB [i] = clipKdTreeA->getNeighbourPositionIndex(0);
		}
		else {
			clipSurfelsOtherB [i] = -1;
		}
	}

	
	
	////////////////////////////////////////////////////////////////
	// compute clip surfels
	//
	for (i = 0; i < nofClipSurfelsA; i++) {
		
		if (clipSurfelsOtherA[i] >= 0) {
			nOfFeatureSurfels++;
		}
	}

	for (i = 0; i < nofClipSurfelsB; i++) {
	
		if ( (clipSurfelsOtherB[i] >= 0) && (clipSurfelsOtherA[clipSurfelsOtherB[i]] != i) ){
			nOfFeatureSurfels++;
		}

	}

	if (nOfFeatureSurfels > 0) {

		if (featurePoints) {
			delete[] featurePoints;
			featurePoints = NULL;
		}
		featurePoints = new FeaturePoint[nOfFeatureSurfels];
		nOfFeatureSurfels = 0;

		for (i = 0; i < nofClipSurfelsA; i++) {
			
			if (clipSurfelsOtherA[i] >= 0) {
				newPosition = this->sampleFeatureSurfelPosition (clipSurfelsA[i]->getPosition(),
						clipSurfelsB[clipSurfelsOtherA[i]]->getPosition(), clipSurfelsA[i]->getNormal(),
						clipSurfelsB[clipSurfelsOtherA[i]]->getNormal());

				featurePoints[nOfFeatureSurfels].position = newPosition;
				featurePoints[nOfFeatureSurfels].surfel1 = clipSurfelsA[i];
				featurePoints[nOfFeatureSurfels].surfel2 = clipSurfelsB[clipSurfelsOtherA[i]];
				featurePoints[nOfFeatureSurfels].isCovered = false;
				nOfFeatureSurfels++;
			}
		}

		for (i = 0; i < nofClipSurfelsB; i++) {
		
			if ( (clipSurfelsOtherB[i] >= 0) && (clipSurfelsOtherA[clipSurfelsOtherB[i]] != i) ){
				newPosition = this->sampleFeatureSurfelPosition (clipSurfelsB[i]->getPosition(),
						clipSurfelsA[clipSurfelsOtherB[i]]->getPosition(), clipSurfelsB[i]->getNormal(),
						clipSurfelsA[clipSurfelsOtherB[i]]->getNormal());

				featurePoints[nOfFeatureSurfels].position = newPosition;
				featurePoints[nOfFeatureSurfels].surfel1 = clipSurfelsA[clipSurfelsOtherB[i]];
				featurePoints[nOfFeatureSurfels].surfel2 = clipSurfelsB[i];
				featurePoints[nOfFeatureSurfels].isCovered = false;
				nOfFeatureSurfels++;
					
			}

		}

		computeFeatureCurves(featurePoints, nOfFeatureSurfels);
					
	}

}

void CSG_Tool::doRefinement() {

	FeatureList			*featureList;
	unsigned int		nOfFeatureSurfels;
	float				overlapFactor;
	bool				isNum;
	SurfelCollection	*surfelsCSG;
	SurfelInterface		*surfel;
	Scene				*scene;

	scene = Scene::getInstance();

	overlapFactor = configurationWidget->OverlapLineEdit->text().toFloat(&isNum);
	if (!isNum || overlapFactor <= 0.0) {
		overlapFactor = 1.0;
		configurationWidget->OverlapLineEdit->setText("1.0");
	}
	surfelsCSG = objectCSG->getSurfelCollection();
	nOfFeatureSurfels = 0;

	for (surfel = surfelsCSG->getFirstSurfel(); surfel = surfelsCSG->getNextSurfel(); ) {
		
		if (surfel->isFlagOn(SurfelInterface::FEATURE)) {
			scene->setSurfelFlagOn(surfel->getAssociatedSurfel(0), SurfelInterface::FEATURE, true, false);
		}
	}

	// remove old feature surfels
	for (surfel = surfelsCSG->getFirstSurfel(); surfel; ) {

		if (surfel->isFlagOn(SurfelInterface::FEATURE)) {
			surfelsCSG->removeCurrentSurfel();
			surfel = surfelsCSG->getCurrentSurfel();
		}
		else {
			surfel = surfelsCSG->getNextSurfel();
		}
	}


	for (featureList = featureLists->first(); featureList; featureList = featureLists->next()) {

		refineCurve(surfelsCSG, featureList, newFeaturesList);
		nOfFeatureSurfels += sampleCurve(surfelsCSG, featureList, overlapFactor);
	}

	recomputeClipSurfels(surfelsCSG, nOfFeatureSurfels);
}



Vector3D CSG_Tool::sampleFeatureSurfelPosition (Vector3D pos1, Vector3D pos2, Vector3D norm1, Vector3D norm2) {


	// compute feature position as intersection of three planes

	// plane1 = tangent plane of vertex vA
	// plane2 = tangent plane of vertex vB
	// plane3 = plane through (pos1 + pos2) / 2 that is orthogonal
	//          to both plane1 and plane2


	Vector3D newPos;

	float d1 = Vector3D::dotProduct (pos1, norm1);
	float d2 = Vector3D::dotProduct (pos2, norm2);
	
	Vector3D normal3 = Vector3D::crossProduct (norm1, norm2);
	float d3 = Vector3D::dotProduct (0.5 * (pos1 + pos2), normal3);
	
	float t1, t2, t3, t4, t5, t6, t7;
	
	
	t1 = norm1[0]*normal3[1];   
	t2 = norm1[0]*norm2[1];
	t3 = normal3[0]*norm1[1];
	t4 = norm2[1]*normal3[0];
	t5 = normal3[1]*norm2[0];
	t6 = norm2[0]*norm1[1];
	t7 = 1/(-norm2[2]*t1+normal3[2]*t2+norm2[2]*t3-norm1[2]*t4+norm1[2]*t5-normal3[2]*t6);
	newPos[0] = t7*(-norm1[1]*normal3[2]*d2+norm1[1]*d3*norm2[2]-norm1[2]*norm2[1]*d3+norm1[2]*normal3[1]*d2-d1*normal3[1]*norm2[2]+d1*norm2[1]*normal3[2]);
	newPos[1] = -t7*(-norm1[0]*normal3[2]*d2+norm1[0]*d3*norm2[2]+normal3[0]*norm1[2]*d2+normal3[2]*norm2[0]*d1-d3*norm2[0]*norm1[2]-normal3[0]*d1*norm2[2]);
	newPos[2] = t7*(-d1*t4+d3*t2-d2*t1+d1*t5+d2*t3-d3*t6);

	return newPos;

}



bool CSG_Tool::checkToolAvailability() const {

	return (Scene::getInstance()->getNofEnabledObjects() > 1);
}

// ***************
// private slots
// ***************

void CSG_Tool::handleObjectListChanged () {

	Scene            *scene;
	Object           *object;
	int				 numOfObjects = 0;
	
	scene = Scene::getInstance();

	// remove old entries
	this->configurationWidget->modelAComboBox->clear();
	this->configurationWidget->modelBComboBox->clear();
		
	for (object = scene->getFirstObject(); object != 0; object = scene->getNextObject()) {
		if (object->isFlagOn (Object::ENABLED)) {
			this->configurationWidget->modelAComboBox->insertItem (object->getName());
			this->configurationWidget->modelBComboBox->insertItem (object->getName());
			numOfObjects++;
		}
	}

	if (numOfObjects < 2) {
		ToolManager::getInstance()->switchToNavigationTool();
		return;
	}
	
	configurationWidget->modelAComboBox->setCurrentItem (0);
	configurationWidget->modelBComboBox->setCurrentItem (1);

	getObjectsFromComboBoxes();

}
	

void CSG_Tool::initSpatialDataStructures () {


	SurfelInterface             *surfel;
	SurfelCollection			*surfelsA,
							    *surfelsB;
	unsigned int                i;



	surfelsA = objectA->getSurfelCollection();
	positionsA = new Vector3D [surfelsA->getNofSurfels()];
	for (i = 0, surfel = surfelsA->getFirstSurfel(); surfel != 0; surfel = surfelsA->getNextSurfel(), i++) {
		positionsA[i] = surfel->getPosition();
	}
	kdTreeA = new KdTree (positionsA, surfelsA->getNofSurfels(), 10);
	surfelsVectorA = objectA->getSurfelCollection()->buildSurfelVector();


	surfelsB = objectB->getSurfelCollection();
	positionsB = new Vector3D [surfelsB->getNofSurfels()];
	for (i = 0, surfel = surfelsB->getFirstSurfel(); surfel != 0; surfel = surfelsB->getNextSurfel(), i++) {
		positionsB[i] = surfel->getPosition();
	}
	kdTreeB = new KdTree (positionsB, surfelsB->getNofSurfels(), 10);
	surfelsVectorB = objectB->getSurfelCollection()->buildSurfelVector();


	handleRebuild();

}



void CSG_Tool::handleOperationChanged() {

	int old_operation = operation;

	if (configurationWidget->a_plus_B_RadioButton->isChecked()) {
		operation = OP_A_union_B;
	}
	else if (configurationWidget->a_times_B_RadioButton->isChecked()) {
		operation = OP_A_intersection_B;
	}
	else if (configurationWidget->a_minus_B_RadioButton->isChecked()) {
		operation = OP_A_minus_B;
	}
	else if (configurationWidget->b_minus_A_RadioButton->isChecked()) {
		operation = OP_B_minus_A;
	}


	// slot is activated twice, so check if something has changed
	if (old_operation != operation) {
		this->computeCSG();
		sampleIntersectionCurves();
		Scene::getInstance()->emitObjectModifiedSignal();
		Scene::getInstance()->emitRenderFrameSignal();
	}
}


void CSG_Tool::handleRebuild() {



	QApplication::setOverrideCursor (Qt::waitCursor);

	maxClustersize = configurationWidget->maxClustersizeSpinBox->value();

	if (clusterTreeA) {
		delete clusterTreeA;
	}
	clusterTreeA = new ClusterTree();
	clusterTreeA->buildTree(surfelsVectorA, maxClustersize);


	if (clusterTreeB) {
		delete clusterTreeB;
	}
	clusterTreeB = new ClusterTree();
	clusterTreeB->buildTree(surfelsVectorB, maxClustersize);

	QApplication::restoreOverrideCursor ();
}



void CSG_Tool::handleApply () {

	Scene* scene;
	SurfelCollection *surfelsCSG;
	SurfelInterface  *surfel;

	scene = Scene::getInstance();

	delete clusterTreeA;
	delete clusterTreeB;
	
	delete [] positionsA;
	delete kdTreeA;

	delete [] positionsB;
	delete kdTreeB;

	delete surfelsVectorA;
	delete surfelsVectorB;

	objectCSG->setFlag (Object::ENABLED, true);
	
	surfelsCSG = objectCSG->getSurfelCollection();
	for (surfel = surfelsCSG->getFirstSurfel(); surfel; ) {

		if (surfel->isFlagOn(SurfelInterface::INVISIBLE)) {
			surfelsCSG->removeCurrentSurfel();
			surfel = surfelsCSG->getCurrentSurfel();
		}
		else {
			surfel = surfelsCSG->getNextSurfel();
		}
	}

	// set capacity of collection to number of surfels
	surfelsCSG->resetCapacity();

	this->disconnect (Scene::getInstance(), SIGNAL (objectRemoved()), this, SLOT (handleObjectListChanged()));
	scene->removeObject (objectA, false);
	scene->removeObject (objectB, true);
	this->connect (Scene::getInstance(), SIGNAL (objectRemoved()), this, SLOT (handleObjectListChanged()));

	objectCSG = 0;

	ToolManager::getInstance()->switchToNavigationTool();
}


void CSG_Tool::handleCancel () {

	ToolManager::getInstance()->switchToNavigationTool();
}




void CSG_Tool::handleComboBoxesUpdated() {

	Object *old_ObjectA, *old_ObjectB;

	old_ObjectA = objectA;
	old_ObjectB = objectB;

	reset();
	getObjectsFromComboBoxes();
	initCSG();
}


void CSG_Tool::getObjectsFromComboBoxes() {


	Object  *object;
	Scene*  scene;
	int     index;
	
	scene = Scene::getInstance();
	index = 0;
	objectA = objectB = NULL;

	for (object = scene->getFirstObject(); object != 0; object = scene->getNextObject()) {
		if (object->isFlagOn (Object::ENABLED)) {
		
			if (index == this->configurationWidget->modelAComboBox->currentItem()) {
				objectA = object;
			}

			if (index == this->configurationWidget->modelBComboBox->currentItem()) {
				objectB = object;
			}
			index++;
		}
	}
}

void CSG_Tool::handlePreviewRendererToggled (bool isOn) {
	usePreviewRenderer = isOn;
}



void CSG_Tool::handleResamplingChanged() {

	this->computeCSG();
	sampleIntersectionCurves();

	Scene::getInstance()->emitObjectModifiedSignal();
	Scene::getInstance()->emitRenderFrameSignal();

}

void CSG_Tool::initCSG() {

	Scene *scene = Scene::getInstance();

	if (scene->getNofObjects() < 2) {
		StatusBar::getInstance()->showMessage("Two objects are needed to perform a CSG operation!");	
		return;
	}
	else if (!objectA) {
		// this should not happen
		StatusBar::getInstance()->showMessage("CSG could not be performed because objectA is invalid!");	
		return;
	}
	else if (!objectB) {
		// this should not happen
		StatusBar::getInstance()->showMessage("CSG could not be performed because objectB is invalid!");	
		return;
	}
	else if (objectA == objectB) {
		// this should not happen
		StatusBar::getInstance()->showMessage("A CSG operation cannot be performed with only one object!");	
		return;
	}
	
	QApplication::setOverrideCursor (Qt::waitCursor);
		
	RendererManager::getInstance()->getPreviewRenderer()->setInteractState (RendererInterface::INTERACT_DYNAMIC);

	clusterTreeA = 0;
	clusterTreeB = 0;

	featureLists = new ObjListDbl<FeatureList*>();	// list of feature lists
	featureLists->setAutoDeleteEnabled(true);	// delete these lists automatically
	newFeaturesList = new FeatureList();			// list for new created feature points
	newFeaturesList->setAutoDeleteEnabled(true);	// these points must be removed when the list is cleared / deleted

	StatusBar::getInstance()->showMessage("Checking and Initializing Models...");	
	if (checkAndInitModel(objectA) && checkAndInitModel(objectB)) {
		
		StatusBar::getInstance()->showMessage("Initializing Datastructures...");	
		initSpatialDataStructures();				
		
		objectA->setFlag (Object::VISIBLE, false);
		objectB->setFlag (Object::VISIBLE, false);
		
		objectCSG = new Object();
		objectCSG->getSurfelCollection()->setPropertyDescriptor (objectA->getSurfelCollection()->getPropertyDescriptor());
		objectCSG->setName ("CSG_object", false);
		objectCSG->setFlag (Object::ENABLED, false);
		
		// perform CSG operation
		StatusBar::getInstance()->showMessage("Perform CSG...");	
		operation = -1;
		handleOperationChanged();
		StatusBar::getInstance()->showMessage("CSG completed");	

		// add CSG object to scene and set it as active object
		this->disconnect (Scene::getInstance(), SIGNAL (objectAdded()), this, SLOT (handleObjectListChanged()));
		scene->addObject (objectCSG, true, true);		
		scene->setActiveObject (objectA->getName());
		this->connect (Scene::getInstance(), SIGNAL (objectAdded()), this, SLOT (handleObjectListChanged()));
		
		setGlobalTransformationActive(false);
		
		this->configurationWidget->applyPushButton->setEnabled (true);
	}
	else {
		objectCSG = NULL;
		this->configurationWidget->applyPushButton->setEnabled (false);	
	}
	
	QApplication::restoreOverrideCursor ();
}


bool CSG_Tool::checkAndInitModel(Object *object) {

	SurfelCollection		*surfels;
	SurfelInterface			*surfel;
	SurfelInterface::Flags	flags, 
							deleteFlags;
	Scene					*scene;

	scene = Scene::getInstance();
	surfels = object->getSurfelCollection();

	if (surfels->getPropertyDescriptor() & SurfelInterface::ASSOCIATED_SURFELS) {
		// has collection already associated surfels?
		for (surfel = surfels->getFirstSurfel(); surfel; surfel = surfels->getNextSurfel()) {
			if (surfel->getNofAssociatedSurfels() > 0) {
				break;
			}
		}

		if (surfel) {
			// Either delete associated surfels or do not use this object for CSG
			QString str;
			str = "Object ";
			str.append(object->getName());
			str.append(" already contains associated surfels.\n");
			str.append("If you use this object for CSG, all associated surfels are deleted.\n");
			str.append("Do you want to proceed with this object?\n\n");
			switch( QMessageBox::warning( configurationWidget, "Associated Surfels", str, "Yes", "No", 0, 0, 1 ) ){

			  case 0: // delete the associated surfels
				  for (; surfel; surfel = surfels->getNextSurfel()) {
					  scene->clearAssociatedSurfels(surfel, false);
				  }
				  break;
			  case 1: // try next object
				  // exit
				  return false;
			  }
		}				
	}

	// delete these flags
	deleteFlags = SurfelInterface::FEATURE | SurfelInterface::CLIP_ORIENTATION | SurfelInterface::CLIP_BOUNDARY_A | SurfelInterface::CLIP_BOUNDARY_B;

	for (surfel = surfels->getFirstSurfel(); surfel; surfel = surfels->getNextSurfel()) {
		flags = surfel->getFlags();
		if (flags & deleteFlags) {
			scene->setSurfelFlags(surfel, flags & (~deleteFlags), false);
		}
	}


	return true;

}

void CSG_Tool::reset() {
	
	if (featureLists) {
		delete featureLists;
		featureLists = NULL;
		delete newFeaturesList;
		newFeaturesList = NULL;
	}
	if (featurePoints) {
		delete[] featurePoints;
		featurePoints = NULL;
	}

	if (objectCSG) {
		delete clusterTreeA;
		clusterTreeA = NULL;
		delete clusterTreeB;
		clusterTreeB = NULL;
		
		delete [] positionsA;
		positionsA = NULL;
		delete kdTreeA;
		kdTreeA = NULL;
		
		delete [] positionsB;
		positionsB = NULL;
		delete kdTreeB;
		kdTreeB = NULL;
		
		delete surfelsVectorA;
		surfelsVectorA = NULL;
		delete surfelsVectorB;
		surfelsVectorB = NULL;
		
		Scene* scene;
		scene = Scene::getInstance();
		
		objectA->setFlag (Object::VISIBLE, true);
		objectB->setFlag (Object::VISIBLE, true);
		
		this->disconnect (Scene::getInstance(), SIGNAL (objectRemoved()), this, SLOT (handleObjectListChanged()));
		scene->removeObject (objectCSG, true);
		this->connect (Scene::getInstance(), SIGNAL (objectRemoved()), this, SLOT (handleObjectListChanged()));
		objectCSG = NULL;
	}

	if (clipPositionsA) {
		delete[] clipPositionsA;
		delete[] clipPositionsB;
		delete[] clipSurfelsA;
		delete[] clipSurfelsB;
		delete[] clipSurfelsOtherA;
		delete[] clipSurfelsOtherB;
		delete   clipKdTreeA;
		delete   clipKdTreeB;
		clipPositionsA		= NULL;
		clipPositionsB		= NULL;
		clipSurfelsA		= NULL;
		clipSurfelsB		= NULL;
		clipSurfelsOtherA	= NULL;
		clipSurfelsOtherB	= NULL;
		clipKdTreeA			= NULL;
		clipKdTreeB			= NULL;
	}

}

// ****************
// access functions
// ****************

ToolInterface *createTool() {
	return new CSG_Tool();
}

const char *getToolName() {
	return "CSG_Tool";
}





	
