// Title:   CSG_Tool.h
// Created: Thu Sep 25 15:07:31 2003
// Authors: Richard Keiser, Mark Pauly
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __CSG_TOOL_H_
#define __CSG_TOOL_H_

#include <qwidget.h>
#include <qtoolbutton.h>
#include <qbuttongroup.h>
#include <qpoint.h>
#include <qpointarray.h>
#include <qcursor.h>
#include <qobject.h>
#include <qimage.h>
#include <qpopmenu.h>
#include <qcombobox.h>
#include <qfiledialog.h>

#include "../../../../DataTypes/src/MyDataTypes.h"
#include "../../../../DataTypes/src/Vector3D.h"
#include "../../../../Utilities/src/Matrix.h"

#include "../../../../Core/RendererManager/src/RendererManager.h"
#include "../../../../Core/Interfaces/src/ToolInterface.h"
#include "../../../../Core/DataStructures/src/kdTree.h"
#include "../../../../Core/DataStructures/src/ClusterTree.h"
#include "../../../../Core/DataStructures/src/ObjListDbl.h"
#include "FeatureList.h"
#include <math.h>

#include "../UI/CSG_ConfigurationWidget.h"


/**
 * Tool for performing boolean operations of two models, i.e. union, intersection and difference.
 *
 * @author Mark Pauly
 * @version 2.0
 */
class CSG_Tool : public ToolInterface {

	Q_OBJECT

public:

	TOOL_API CSG_Tool();
	TOOL_API virtual ~CSG_Tool();

	// ***********************
	// interface ToolInterface
	// ***********************

	TOOL_API QToolButton *addButtonToToolBar (QToolBar *toolBar);	
	TOOL_API QCursor *getToolCursor() const;
	TOOL_API QToolButton *getToolButton();
	TOOL_API QWidget *getToolConfigurationWidget() const;
	TOOL_API bool isAvailable() const;

	TOOL_API void handleMousePressEvent   (QMouseEvent *me);
	TOOL_API void handleKeyPressEvent	  (QKeyEvent   *ke);
	TOOL_API void handleMouseReleaseEvent (QMouseEvent *me);
	TOOL_API void handleMouseMoveEvent    (QMouseEvent *me);
	TOOL_API void handleKeyReleaseEvent   (QKeyEvent   *ke);
	TOOL_API void selectTool();
	TOOL_API void deselectTool();

	void setGlobalTransformationActive(bool isOn);



public slots:

	void handleRendererWidgetRepainted();

private:

	QPoint					lastMouseDragPosition;
	QCursor                 *toolCursor;
	QCursor                 *standardCursor,
							*transformCursor;
	QToolButton             *csg_Button;	

	int						operation;

	// the configuration widget
	CSG_ConfigurationWidget *configurationWidget;

	bool                    usePreviewRenderer;


	// objects used in CSG
	Object					 *objectA,
		                     *objectB,
							 *objectCSG;

	/////////////////////////////////////////////////////////////////////////////////////
	// temporary pointers ("global" variables") so function calls don't need arguments
	Object                   *object1,
		                     *object2;
	ClusterTree				 *clusterTree;
	KdTree					 *kdTree;
	QVector<SurfelInterface> *surfelsVector;
	SurfelInterface::Flag    clipFlag;	
	MyDataTypes::TransformationMatrix16f objectToObjectTransformation;
	float					 objectToObjectScaleFactor;
	/////////////////////////////////////////////////////////////////////////////////////


	KdTree					 *kdTreeA,
		                     *kdTreeB,
							 *clipKdTreeA,
							 *clipKdTreeB; 

	ClusterTree				 *clusterTreeA,
		                     *clusterTreeB;

	Vector3D                 *positionsA,
		                     *positionsB,
							 *clipPositionsA,
							 *clipPositionsB;

	QVector<SurfelInterface> *surfelsVectorA,
		                     *surfelsVectorB;

	SurfelInterface			 **clipSurfelsA,
							 **clipSurfelsB;
	int						 *clipSurfelsOtherA,
							 *clipSurfelsOtherB;

	bool                     clipOrientation;

	int						 maxClustersize;

	int					     numTotalQueries;
	int	             		 numBoundarySurfels;

	ObjListDbl<FeatureList*> *featureLists;
	FeatureList				 *newFeaturesList;
	FeaturePoint			 *featurePoints;
	

	void computeCSG();
	void classifySurfels ();
	void classifyNode (ClusterNode *node);
	void sampleIntersectionCurves();
	void computeFeatureCurves(FeaturePoint *featurePoints, int nOfFeaturePoints);
	void subsampleCurve(FeatureList *featureList, const float minDistFactor);
	void refineAngle(SurfelCollection *surfelsCSG, FeatureList *featureList, const float maxAngle, FeatureList *newFeatureList);
	void refineCurve(SurfelCollection *surfelsCSG, FeatureList *featureList, const float maxDistFactor, FeatureList *newFeatureList);
	void refineCurve(SurfelCollection *surfelsCSG, FeatureList *featureList, FeatureList *newFeatureList);
	unsigned int sampleCurve(SurfelCollection *surfelsCSG, FeatureList *featureList, float overlappingFactor);
	void recomputeClipSurfels(SurfelCollection *surfels, unsigned int nOfFeatureSurfels);
	void doRefinement();
	void computeSurfelProperties(SurfelInterface *newSurfel, Vector3D &newPosition, float radius, KdTree *kdTree, unsigned int nOfNeighbors, SurfelInterface **surfels, Scene *scene);
	/**
	 * Computes the intersection curve including the start surfel
	 * The surfels of the intersection curve are returned as a list with ordered surfels.
	 * Note that the <code>startSurfel</startSurfel> is not necessary the start surfel of the list.
	 * If the intersection curve closes a full cycle (it is assumed that there is no junction),
	 * the first and last element of the list will be the same
	 */
	FeatureList* computeCurve(FeaturePoint* startPoint, FeaturePoint *featurePoints, KdTree *featureKdTree);
	//void adjustClipSurfelShape (SurfelInterface *surfel, SurfelInterface *clipSurfel);
	Vector3D sampleFeatureSurfelPosition (Vector3D pos1, Vector3D pos2, Vector3D norm1, Vector3D norm2);
	void getObjectsFromComboBoxes();
	void initSpatialDataStructures();
	bool checkToolAvailability() const;
	
	inline QRgb mixColor(const QRgb color1, const QRgb color2, const float fac) const;
	bool adjustClipSurfel(SurfelInterface *clipSurfel, SurfelInterface *surfel, Scene *scene);
	SurfelInterface* insertNewSurfel(SurfelCollection *surfelsCSG, Vector3D &pos, SurfelInterface *surfel1, SurfelInterface *surfel2, float fac, SurfelInterface::Flag flag);

	void initCSG();
	bool checkAndInitModel(Object *object);
	void reset();

private slots:

	void handleObjectListChanged();
	void handleOperationChanged();
	void handleApply();
	void handleCancel();
	void handleComboBoxesUpdated();
	void handlePreviewRendererToggled (bool isOn);
	void handleRebuild ();
	void handleResamplingChanged();

};

inline QRgb CSG_Tool::mixColor(const QRgb color1, const QRgb color2, const float fac) const{
	
	QColor col1 = QColor(color1);
	QColor col2 = QColor(color2);
	return QColor(  col1.red()*fac + col2.red()*(1.0-fac),
					col1.green()*fac + col2.green()*(1.0-fac),
					col1.blue()*fac + col2.blue()*(1.0-fac)
				 ).rgb();
}


// ****************
// access functions
// ****************

// extern "C" is needed so those functions can be accessed
// with GetProcAddress() (WIN32) and dlsym() (UNIX) respective
extern "C" {
	TOOL_API ToolInterface *createTool();
	TOOL_API const char *getToolName();
}

#endif  // __CSG_TOOL_H_
